import toml 
from ._dataframe_utils import DataFrameType, to_prefered_type
from . import _glaciers_python
from glaciers import get_config

def update_abi_db(abi_db_path: str | None = None, 
                 abi_folder_path: str | None = None) -> DataFrameType:
    """Updates the ABI DB file with new ABIs from the specified folder.

    Args:
        abi_db_path (str | None, optional): Path to the ABI database file. If None,
            uses the path set in the config file. Defaults to None.
        abi_folder_path (str | None, optional): Path to the folder containing ABI files.
            If None, uses the path set in the config file. Defaults to None.

    Returns:
        DataFrameType: A DataFrame containing the updated ABI items.

    Note:
        The ABI DB file is updated with the new ABIs from the specified folder.

    Examples:
        ```python
        # Update using default paths from config
        df = update_abi_db()
        
        # Update using custom paths
        df = update_abi_db("ABIs/ethereum__events__abis.parquet", "ABIs/abi_database")
        ```
    """
    if abi_db_path is None:
        abi_db_path = toml.loads(get_config())["main"]["events_abi_db_file_path"]
    if abi_folder_path is None:
        abi_folder_path = toml.loads(get_config())["main"]["abi_folder_path"]
    df = _glaciers_python.update_abi_db(abi_db_path, abi_folder_path)
    return to_prefered_type(df)

def read_new_abi_folder(abi_folder_path: str | None = None) -> DataFrameType:
    """Reads all ABI files from a specified folder.

    Args:
        abi_folder_path (str | None, optional): Path to the folder containing ABI files.
            If None, uses the path set in the config file. Defaults to None.

    Returns:
        DataFrameType: A DataFrame containing the parsed ABI items from all files
            in the specified folder.

    Examples:
        ```python
        # Read from default folder specified in config
        df = read_new_abi_folder()
        
        # Read from custom folder
        df = read_new_abi_folder("ABIs/abi_database")
        ```
    """
    if abi_folder_path is None:
        abi_folder_path = toml.loads(get_config())["main"]["abi_folder_path"]
        print(f"Reading ABI DataFrame from {abi_folder_path}")
    df = _glaciers_python.read_new_abi_folder(abi_folder_path)
    return to_prefered_type(df)

def read_new_abi_file(path: str) -> DataFrameType:
    """Reads and parses a single ABI file.

    Args:
        path (str): Path to the ABI JSON file.

    Returns:
        DataFrameType: A DataFrame containing the parsed ABI items from the file.

    Examples:
        ```python
        df = read_new_abi_file("ABIs/abi_database/0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2.json")
        ```
    """
    df = _glaciers_python.read_new_abi_file(path)
    return to_prefered_type(df)

def read_new_abi_json(abi: str, address: str) -> DataFrameType:
    """Reads and parses an ABI from a JSON string.

    Args:
        abi (str): The ABI JSON string to parse.
        address (str): The contract address associated with this ABI.

    Returns:
        DataFrameType: A DataFrame containing the parsed ABI items.

    Examples:
        ```python
        abi_json = '{"inputs": [], "name": "totalSupply", ...}'
        address = "0x123..."
        df = read_new_abi_json(abi_json, address)
        ```
    """
    df = _glaciers_python.read_new_abi_json(abi, address)
    return to_prefered_type(df)


import os
from os.path import dirname
import polars as pl
import pandas as pd
import toml
from io import StringIO
import pytest
from glaciers import (
    decode_df,
    decode_df_with_abi_df,
    decode_df_using_single_contract,
    get_config,
    set_config_toml,
    decode_folder,
    decode_file
)

@pytest.fixture
def setup_paths(tmp_path):
    # Get the project directory
    python_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    project_dir = dirname(dirname(python_dir))+"/"
    
    # Create temporary directories
    tmp_logs_dir = tmp_path / "logs"
    tmp_traces_dir = tmp_path / "traces"
    tmp_abi_dir = tmp_path / "abi"
    
    for dir in [tmp_logs_dir, tmp_traces_dir, tmp_abi_dir]:
        dir.mkdir()
    
    # Set config and get paths
    set_config_toml(project_dir+"glaciers_config_edit_example.toml")
    config = get_config()
    config = toml.load(StringIO(config))
    
    # Copy real files to temporary locations
    real_logs_dir = project_dir + config['main']['raw_logs_folder_path']
    real_traces_dir = project_dir + config['main']['raw_traces_folder_path']
    real_events_abi = project_dir + config['main']['events_abi_db_file_path']
    real_functions_abi = project_dir + config['main']['functions_abi_db_file_path']
    
    # Copy the first log and trace files
    if os.listdir(real_logs_dir):
        sample_log = os.path.join(real_logs_dir, os.listdir(real_logs_dir)[0])
        pl.read_parquet(sample_log).write_parquet(tmp_logs_dir / "sample_log.parquet")
    
    if os.listdir(real_traces_dir):
        sample_trace = os.path.join(real_traces_dir, os.listdir(real_traces_dir)[0])
        pl.read_parquet(sample_trace).write_parquet(tmp_traces_dir / "sample_trace.parquet")
    
    # Copy ABI files
    pl.read_parquet(real_events_abi).write_parquet(tmp_abi_dir / "events_abi.parquet")
    pl.read_parquet(real_functions_abi).write_parquet(tmp_abi_dir / "functions_abi.parquet")
    
    return {
        'events_abi_path': str(tmp_abi_dir/"events_abi.parquet"),
        'functions_abi_path': str(tmp_abi_dir/"functions_abi.parquet"),
        'logs_folder_path': str(tmp_logs_dir),
        'traces_folder_path': str(tmp_traces_dir)
    }

@pytest.fixture
def sample_logs_df(setup_paths):
    # Read the sample log file from the temporary folder
    logs_file = os.path.join(setup_paths['logs_folder_path'], "sample_log.parquet")
    return pl.read_parquet(logs_file)

@pytest.fixture
def sample_traces_df(setup_paths):
    # Read the sample trace file from the temporary folder
    traces_file = os.path.join(setup_paths['traces_folder_path'], "sample_trace.parquet")
    return pl.read_parquet(traces_file)

@pytest.fixture
def sample_events_abi_df(setup_paths):
    # Read the ABI DataFrame from the temporary parquet file
    return pl.read_parquet(setup_paths['events_abi_path'])

@pytest.fixture
def sample_functions_abi_df(setup_paths):
    # Read the ABI DataFrame from the temporary parquet file
    return pl.read_parquet(setup_paths['functions_abi_path'])



def test_decode_folder(setup_paths):
    # Test logs decoding
    decode_folder(
        decoder_type="log", 
        abi_db_path=setup_paths['events_abi_path'], 
        folder_path=setup_paths['logs_folder_path']
    )
    
    # Verify that decoded files were created
    decoded_folder = os.path.join(os.path.dirname(setup_paths['logs_folder_path']), "decoded")
    assert os.path.exists(decoded_folder)
    assert len(os.listdir(decoded_folder)) > 0
    
    # Test traces decoding
    decode_folder(
        decoder_type="trace", 
        abi_db_path=setup_paths['functions_abi_path'], 
        folder_path=setup_paths['traces_folder_path']
    )
    
    # Verify that decoded files were created
    decoded_folder = os.path.join(os.path.dirname(setup_paths['traces_folder_path']), "decoded")
    assert os.path.exists(decoded_folder)
    assert len(os.listdir(decoded_folder)) > 0

def test_decode_file(setup_paths):
    # Test log file decoding
    log_file = os.path.join(setup_paths['logs_folder_path'], os.listdir(setup_paths['logs_folder_path'])[0])
    decoded_logs = decode_file(
        decoder_type="log",
        file_path=log_file,
        abi_db_path=setup_paths['events_abi_path']
    )
    
    assert isinstance(decoded_logs, pl.DataFrame)
    assert not decoded_logs.is_empty()
    for col in ["event_keys", "event_values", "event_json"]:
        assert col in decoded_logs.columns
    
    # Test trace file decoding
    trace_file = os.path.join(setup_paths['traces_folder_path'], os.listdir(setup_paths['traces_folder_path'])[0])
    decoded_traces = decode_file(
        decoder_type="trace",
        file_path=trace_file,
        abi_db_path=setup_paths['functions_abi_path']
    )
    
    assert isinstance(decoded_traces, pl.DataFrame)
    assert not decoded_traces.is_empty()
    for col in ["input_keys", "input_values", "input_json", "output_keys", "output_values", "output_json"]:
        assert col in decoded_traces.columns 

def test_decode_df(sample_logs_df, sample_traces_df, setup_paths):    
    result = decode_df("log", sample_logs_df, abi_db_path=setup_paths['events_abi_path'])
    assert isinstance(result, (pl.DataFrame, pd.DataFrame))
    assert not result.is_empty()
    assert "name" in result.columns
    assert "Transfer" in result["name"].to_list()
    for col in ["event_keys", "event_values", "event_json"]:
        assert col in result.columns
    
    result = decode_df("trace", sample_traces_df, abi_db_path=setup_paths['functions_abi_path'])
    assert isinstance(result, (pl.DataFrame, pd.DataFrame))
    assert not result.is_empty()
    assert "name" in result.columns
    assert "transfer" in result["name"].to_list()
    for col in ["input_keys", "input_values", "input_json", "output_keys", "output_values", "output_json"]:
        assert col in result.columns

def test_decode_df_with_abi_df(sample_logs_df, sample_traces_df, sample_events_abi_df, sample_functions_abi_df):
    result = decode_df_with_abi_df("log", sample_logs_df, sample_events_abi_df)
    assert isinstance(result, (pl.DataFrame, pd.DataFrame))
    assert not result.is_empty()
    assert "name" in result.columns
    assert "Transfer" in result["name"].to_list()
    for col in ["event_keys", "event_values", "event_json"]:
        assert col in result.columns

    result = decode_df_with_abi_df("trace", sample_traces_df, sample_functions_abi_df)
    assert isinstance(result, (pl.DataFrame, pd.DataFrame))
    assert not result.is_empty()
    assert "name" in result.columns
    assert "transfer" in result["name"].to_list()
    for col in ["input_keys", "input_values", "input_json", "output_keys", "output_values", "output_json"]:
        assert col in result.columns

def test_decode_df_using_single_contract(sample_logs_df, sample_traces_df):
    result = decode_df_using_single_contract(
        "log",
        sample_logs_df,
        "0xE672E0E0101A7F58d728751E2a5e6Da5Ff1FDa64"
    )
    assert isinstance(result, (pl.DataFrame, pd.DataFrame))
    assert not result.is_empty()
    for col in ["event_keys", "event_values", "event_json"]:
        assert col in result.columns

    result = decode_df_using_single_contract(
        "trace",
        sample_traces_df,
        "0xE672E0E0101A7F58d728751E2a5e6Da5Ff1FDa64"
    )
    assert isinstance(result, (pl.DataFrame, pd.DataFrame))
    assert not result.is_empty()
    for col in ["input_keys", "input_values", "input_json", "output_keys", "output_values", "output_json"]:
        assert col in result.columns

def test_invalid_decoder_type(sample_logs_df):
    with pytest.raises(ValueError, match="Decoder type must be one of"):
        decode_df("invalid_type", sample_logs_df)


import os
from os.path import dirname
import polars as pl
import glaciers as gl
import toml
from io import StringIO

# Get the project directory
python_dir = os.path.dirname(os.path.abspath(__file__))
project_dir =  dirname(dirname(python_dir))+"/"
print(f"Project dir: {project_dir}")

# Set and get the configs
gl.set_config("decoder.decoded_chunk_size", 500_000)
gl.set_config_toml(project_dir+"glaciers_config_edit_example.toml")
config = gl.get_config()
print(f"Glaciers config:\n{config}")
config = toml.load(StringIO(config))

# Instantiate the paths to the files and folders
events_abi_file_path = project_dir+config['main']['events_abi_db_file_path']
functions_abi_file_path = project_dir+config['main']['functions_abi_db_file_path']
abi_folder_path = project_dir+config['main']['abi_folder_path']
logs_folder_path = project_dir+config['main']['raw_logs_folder_path']
traces_folder_path = project_dir+config['main']['raw_traces_folder_path']


######## Test update_abi_db ########
# Reads ABIs in a folder and append to the abi parquet file
#
# This function loads each ABI definitions from a ABI folder and append the new itens (events and functions, according to the configs)
# into an existing abi DF (parquet file) the new unique entries. Function also output the dataframe.
#
# # Arguments
# - `abi_db_path`: Optional, default in config. Path to the parquet file containing the existing DataFrame.
# - `abi_folder_path`: Optional, default in config. Path to the folder containing ABI JSON files
#
# # Returns
# A `PyResult` containing a `PyDataFrame` with all functions and events itens signatures
#
# # Errors
# Returns a `PyValueError` if there are issues reading or processing the ABIs
abis_df = gl.update_abi_db(abi_db_path=events_abi_file_path, abi_folder_path=abi_folder_path)
print(f"\nFirst 5 rows of updatedABIs DataFrame:\n{abis_df.head()}\n\n")


######## Test read_new_abi_folder ########
# Reads ABIs in a folder
#
# This function loads ABI files from a folder and creates a DataFrame containing functions 
# and events itens (according to the configs) found in the ABI files. It doesn't save them in a DB file.
#
# # Arguments
# - `abi_folder_path`: Optional, default in config. Path to the folder containing ABI JSON files
#
# # Returns
# A `PyResult` containing a `PyDataFrame` with all functions and events itens signatures
#
# # Errors
# Returns a `PyValueError` if there are issues reading or processing the ABIs
folder_df = gl.read_new_abi_folder(abi_folder_path=abi_folder_path)
print(f"\nABIs DataFrame from folder:\n{folder_df.head()}\n\n")


######## Test read_new_abi_file ########
# Reads a single ABI file
#
# This function loads ABI definitions from a file and creates a DataFrame containing functions 
# and events itens (according to the configs) found in the ABI file.
#
# # Arguments
# - `path`: Path to the ABI JSON file
#
# # Returns
# A `PyResult` containing a `PyDataFrame` with all functions and events itens signatures
#
# # Errors
# Returns a `PyValueError` if there are issues reading or processing the ABI
abi_file = os.path.join(abi_folder_path, os.listdir(abi_folder_path)[0])  # Get first ABI file
file_df = gl.read_new_abi_file(path=abi_file)
print(f"\nABIs DataFrame from single file:\n{file_df.head()}\n\n")


######## Test read_new_abi_json ########
# Reads a single ABI JSON string
#
# This function loads ABI definitions from a JSON string and creates a DataFrame containing
# functions and events itens (according to the configs) found in the ABI.
#
# # Arguments
# - `abi`: ABI JSON string
# - `address`: Contract address as string
#
# # Returns
# A `PyResult` containing a `PyDataFrame` with all functions and events
#
# # Errors
# Returns a `PyValueError` if there are issues reading or processing the ABI
with open(abi_file, 'r') as f:
    abi_json = """
        [{
            "anonymous": false,
            "inputs": [
                {
                    "indexed": true,
                    "internalType": "address",
                    "name": "from",
                    "type": "address"
                },
                {
                    "indexed": true,
                    "internalType": "address",
                    "name": "to",
                    "type": "address"
                },
                {
                    "indexed": false,
                    "internalType": "uint256",
                    "name": "value",
                    "type": "uint256"
                }
            ],
            "name": "Transfer",
            "type": "event"
        }]
    """
    address = "0xE672E0E0101A7F58d728751E2a5e6Da5Ff1FDa64"
    item_df = gl.read_new_abi_json(abi_json, address)
    print(f"\nABIs DataFrame from JSON string:\n{item_df.head()}\n\n")



######## Test decode_folder ########
# Decode a folder of logs/traces.
#
# This function takes a logs folder path and a abi parquet file path. It iterate through 
# logs files, decode, and save them into decoded logs' files.
#
# # Arguments
# - `decoder_type`: Type of the decoder to use, allowed values = ["log", "trace"]
# - `abi_db_path`: Optional, default in config. Path to the abi parquet file
# - `folder_path`: Optional, default in config. Path to the folder containing the logs/traces files
#
# # Returns
# No Return
#
# # Errors
# Returns a `PyValueError` if there are issues processing the logs
gl.decode_folder(decoder_type="log", abi_db_path=events_abi_file_path, folder_path=logs_folder_path)
print(f"\n Decoded logs saved in the decoded folder.\n\n")


######## Test decode_file ########
# Decode a log/trace file
#
# This function takes a log file path and a abi db parquet file and decode the file
# to a decoded logs/traces DataFrame.
#
# # Arguments
# - `decoder_type`: Type of the decoder to use, allowed values = ["log", "trace"]
# - `file_path`: Optional, default in config. Path to the log/trace file
# - `abi_db_path`: Optional, default in config. Path to the abi parquet file
#
# # Returns
# A `PyResult` containing a decoded logs/traces DataFrame or an error
trace_file = os.path.join(traces_folder_path, os.listdir(traces_folder_path)[0])  # Get first trace file
decoded_df = gl.decode_file(decoder_type="trace", file_path=trace_file, abi_db_path=functions_abi_file_path)
print(f"\nDecoded Traces in the trace file {trace_file}:\n{decoded_df.head()}\n\n")


######## Test decode_df ########
# Decode a logs/traces DataFrame
#
# This function takes a raw logs/traces DataFrame and a abi DB parquet file path and decode the df
# to a decoded logs/traces DataFrame.
#
# # Arguments
# - `df`: A DataFrame containing raw blockchain logs/traces
# - `decoder_type`: Type of the decoder to use, allowed values = ["log", "trace"]
# - `abi_db_path`: Optional, default in config. Path to the abi parquet file
#
# # Returns
# A `PyResult` containing a decoded logs/traces DataFrame or an error
#
# # Errors
# Returns a `PyValueError` if there are issues processing the logs

logs_df_path = os.path.join(logs_folder_path, os.listdir(logs_folder_path)[0])  # Get first logs file
logs_df = pl.read_parquet(logs_df_path)
decoded_df = gl.decode_df(df=logs_df, decoder_type="log", abi_db_path=events_abi_file_path)
print(f"\nDecoded Logs DataFrame:\n{decoded_df.head()}\n\n")


######## Test decode_df_with_abi_df ########
# Decode a logs/traces DataFrame with a abi DataFrame
#
# This function takes a raw logs/traces DataFrame and a abi DataFrame and decode the df
# to a decoded logs/traces DataFrame.
#
# # Arguments
# - `df`: A DataFrame containing raw blockchain logs/traces
# - `abi_df`: A DataFrame containing the ABI definitions
# - `decoder_type`: Type of the decoder to use, allowed values = ["log", "trace"]
#
# # Returns
# A `PyResult` containing a decoded logs/traces DataFrame or an error
logs_df_path = os.path.join(logs_folder_path, os.listdir(logs_folder_path)[0])  # Get first logs file
logs_df = pl.read_parquet(logs_df_path)
abi_df = pl.read_parquet(events_abi_file_path)
decoded_df = gl.decode_df_with_abi_df(df=logs_df, abi_df=abi_df, decoder_type="log")
print(f"\nDecoded Logs using ABI DataFrame:\n{decoded_df.head()}\n\n")


######## Test unnest_event ########
# Filter and unnest an event from a decoded logs' DataFrame. After filtering using the combination of the
# optional arguments, the event needs to be unique.
#
# Arguments
# - `decoded_log_df`: A DataFrame containing the decoded logs
# - `event_name`: Optional, default None. The name of the event to unnest
# - `full_signature`: Optional, default None. The full signature of the event to unnest
# - `event_address`: Optional, default None. The address of the event to unnest
# - `topic0`: Optional, default None. The topic0 of the event to unnest
#
# Returns
# A `PyResult` containing a unnested event's `PyDataFrame` or an error
#
# Errors
# Returns a `PyValueError` if there are issues processing the unnesting or the event is not unique.

unnested_df = gl.unnest_event(decoded_df, full_signature='event Transfer(address indexed from, address indexed to, uint256 value)')
print(f"\nUnnested Logs DataFrame:\n{unnested_df.head(5)}\n\n")


import pytest
import toml
from glaciers import get_config, set_config, set_config_toml
import os

def test_default_config():
    """Test getting default configuration values"""
    config = get_config()
    config = toml.loads(config)
    # Test getting some default values
    assert config["glaciers"]["preferred_dataframe_type"] == "Polars"
    assert config["glaciers"]["unnesting_hex_string_encoding"] == False
    assert config["main"]["events_abi_db_file_path"] == "ABIs/ethereum__events__abis.parquet"
    assert config["main"]["functions_abi_db_file_path"] == "ABIs/ethereum__functions__abis.parquet"
    assert config["main"]["abi_folder_path"] == "ABIs/abi_database"
    assert config["main"]["raw_logs_folder_path"] == "data/logs"
    assert config["main"]["raw_traces_folder_path"] == "data/traces"
    assert config["abi_reader"]["abi_read_mode"] == "Events"
    assert config["abi_reader"]["unique_key"] == ["hash", "full_signature", "address"]
    assert config["decoder"]["algorithm"] == "Hash"
    assert config["decoder"]["output_hex_string_encoding"] == False
    assert config["decoder"]["output_file_format"] == "parquet"
    assert config["decoder"]["max_concurrent_files_decoding"] == 16
    assert config["decoder"]["max_chunk_threads_per_file"] == 16
    assert config["decoder"]["decoded_chunk_size"] == 500000
    assert config["log_decoder"]["log_schema"]["log_alias"] == {"topic0": "topic0", "topic1": "topic1", "topic2": "topic2", "topic3": "topic3", "data": "data", "address": "address"}
    assert config["log_decoder"]["log_schema"]["log_datatype"] == {"topic0": "Binary", "topic1": "Binary", "topic2": "Binary", "topic3": "Binary", "data": "Binary", "address": "Binary"}
    assert config["trace_decoder"]["trace_schema"]["trace_alias"] == {"selector": "selector", "action_input": "action_input", "result_output": "result_output", "action_to": "action_to"}
    assert config["trace_decoder"]["trace_schema"]["trace_datatype"] == {"selector": "Binary", "action_input": "Binary", "result_output": "Binary", "action_to": "Binary"}

def test_set_config_toml():
    """Test loading configuration from TOML file"""
    # Create a temporary test config file
    test_config = """
    [glaciers]
    preferred_dataframe_type = "pandas"
    unnesting_hex_string_encoding = true

    [decoder]
    algorithm = "hash_address"
    max_concurrent_files_decoding = 8
    """
    
    with open("test_config.toml", "w") as f:
        f.write(test_config)

    # Load the test config
    set_config_toml("test_config.toml")
    config = get_config()
    config = toml.loads(config)
    # Verify the values were updated
    assert config["glaciers"]["preferred_dataframe_type"] == "Pandas"
    assert config["glaciers"]["unnesting_hex_string_encoding"] == True
    assert config["decoder"]["algorithm"] == "HashAddress"
    assert config["decoder"]["max_concurrent_files_decoding"] == 8

    # Clean up
    os.remove("test_config.toml")

def test_set_config():
    """Test setting individual configuration values"""
    set_config("glaciers.preferred_dataframe_type", "polars")
    set_config("glaciers.preferred_dataframe_type", "pandas")
    set_config("glaciers.unnesting_hex_string_encoding", False)
    set_config("glaciers.unnesting_hex_string_encoding", True)
    set_config("glaciers.unnesting_hex_string_encoding", 0)
    set_config("glaciers.unnesting_hex_string_encoding", 1)
    set_config("main.events_abi_db_file_path", "ABIs/ethereum__abis.parquet")
    set_config("main.functions_abi_db_file_path", "ABIs/ethereum__abis.parquet")
    set_config("main.abi_folder_path", "ABIs")
    set_config("main.raw_logs_folder_path", "data")
    set_config("main.raw_traces_folder_path", "data")
    set_config("abi_reader.abi_read_mode", "Events")
    set_config("abi_reader.abi_read_mode", "Functions")
    set_config("abi_reader.abi_read_mode", "Both")
    set_config("abi_reader.output_hex_string_encoding", False)
    set_config("abi_reader.output_hex_string_encoding", True)
    set_config("abi_reader.output_hex_string_encoding", 0)
    set_config("abi_reader.output_hex_string_encoding", 1)
    set_config("abi_reader.unique_key", ["hash", "full_signature", "address"])
    set_config("abi_reader.unique_key", ["hash", "full_signature"])
    set_config("abi_reader.unique_key", "hash")
    set_config("abi_reader.unique_key", "full_signature")
    set_config("abi_reader.unique_key", "address")
    set_config("decoder.algorithm", "Hash")
    set_config("decoder.algorithm", "Hash_Address")
    set_config("decoder.output_hex_string_encoding", False)
    set_config("decoder.output_hex_string_encoding", True)
    set_config("decoder.output_hex_string_encoding", 0)
    set_config("decoder.output_hex_string_encoding", 1)
    set_config("decoder.output_file_format", "parquet")
    set_config("decoder.output_file_format", "csv")
    set_config("decoder.max_concurrent_files_decoding", 1)
    set_config("decoder.max_chunk_threads_per_file", 1)
    set_config("decoder.decoded_chunk_size", 1)
    set_config("log_decoder.log_schema.log_alias.topic0", "t0")
    set_config("log_decoder.log_schema.log_alias.topic1", "t1")
    set_config("log_decoder.log_schema.log_alias.topic2", "t2")
    set_config("log_decoder.log_schema.log_alias.topic3", "t3")
    set_config("log_decoder.log_schema.log_alias.data", "d")
    set_config("log_decoder.log_schema.log_alias.address", "event_address")
    set_config("log_decoder.log_schema.log_datatype.topic0", "HexString")
    set_config("log_decoder.log_schema.log_datatype.topic1", "HexString")
    set_config("log_decoder.log_schema.log_datatype.topic2", "HexString")
    set_config("log_decoder.log_schema.log_datatype.topic3", "HexString")
    set_config("log_decoder.log_schema.log_datatype.data", "HexString")
    set_config("log_decoder.log_schema.log_datatype.address", "HexString")
    set_config("trace_decoder.trace_schema.trace_alias.selector", "4bytes")
    set_config("trace_decoder.trace_schema.trace_alias.action_input", "input")
    set_config("trace_decoder.trace_schema.trace_alias.result_output", "output")
    set_config("trace_decoder.trace_schema.trace_alias.action_to", "to")
    set_config("trace_decoder.trace_schema.trace_datatype.selector", "HexString")
    set_config("trace_decoder.trace_schema.trace_datatype.action_input", "HexString")
    set_config("trace_decoder.trace_schema.trace_datatype.result_output", "HexString")
    set_config("trace_decoder.trace_schema.trace_datatype.action_to", "HexString")
    expected_config = '''
        [glaciers]
        preferred_dataframe_type = "Pandas"
        unnesting_hex_string_encoding = true

        [main]
        events_abi_db_file_path = "ABIs/ethereum__abis.parquet"
        functions_abi_db_file_path = "ABIs/ethereum__abis.parquet"
        abi_folder_path = "ABIs"
        raw_logs_folder_path = "data"
        raw_traces_folder_path = "data"

        [abi_reader]
        abi_read_mode = "Both"
        output_hex_string_encoding = true
        unique_key = ["address"]

        [decoder]
        algorithm = "HashAddress"
        output_hex_string_encoding = true
        output_file_format = "csv"
        max_concurrent_files_decoding = 1
        max_chunk_threads_per_file = 1
        decoded_chunk_size = 1

        [log_decoder.log_schema]
        log_alias = { topic0 = "t0", topic1 = "t1", topic2 = "t2", topic3 = "t3", data = "d", address = "event_address" }
        log_datatype = { topic0 = "HexString", topic1 = "HexString", topic2 = "HexString", topic3 = "HexString", data = "HexString", address = "HexString" }

        [trace_decoder.trace_schema]
        trace_alias = { selector = "4bytes", action_input = "input", result_output = "output", action_to = "to" }
        trace_datatype = { selector = "HexString", action_input = "HexString", result_output = "HexString", action_to = "HexString" }
    '''
    expected_config = toml.loads(expected_config)
    config = get_config()
    config = toml.loads(config)
    print(config)
    print(expected_config)
    assert config == expected_config

def test_invalid_config():
    """Test error handling for invalid configurations"""
    # Test invalid value type
    with pytest.raises(ValueError):
        set_config("decoder.max_concurrent_files_decoding", "invalid")
        config = get_config()
        config = toml.loads(config)
        assert config["decoder"]["max_concurrent_files_decoding"] == "invalid"

    # Test invalid dataframe type
    with pytest.raises(ValueError):
        set_config("glaciers.preferred_dataframe_type", "invalid")
        config = get_config()
        config = toml.loads(config)
        assert config["glaciers"]["preferred_dataframe_type"] == "invalid"



import pytest
import polars as pl
import pandas as pd
from glaciers import unnest_event, unnest_trace

@pytest.fixture
def sample_decoded_df():
    return pl.DataFrame({
        "log_index": [18, 155],
        "transaction_index": [18, 86],
        "block_number": [18426296, 18426286],
        "transaction_hash": [bytes.fromhex("96591a6a4ea24492a745464a72e1520774d2f543ea28126d274705a1726c4021"),
                             bytes.fromhex("b94c76f0b52be1e077dddcd759a4bba7fc8976d21bd6ae0f3401cc6bedc1c263")],
        "address": [bytes.fromhex("c02aaa39b223fe8d0a0e5c4f27ead9083c756cc2"),
                   bytes.fromhex("dac17f958d2ee523a2206206994597c13d831ec7")],
        "topic0": [bytes.fromhex("ddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef")] * 2,
        "topic1": [bytes.fromhex("000000000000000000000000eedff72a683058f8ff531e8c98575f920430fdc5"),
                   bytes.fromhex("0000000000000000000000009c0fdf1e074a8db4b3bfc78cf4d48a5518b7a0ba")],
        "topic2": [bytes.fromhex("0000000000000000007a250d5630b4cf539739df2c5dacb4c659f2488d"),
                  bytes.fromhex("000000000000000000a9d1e08c7793af67e9d92fe308d5697fb81d3e43")],
        "topic3": [bytes.fromhex("000000000000000000000000000000000000000002bbedcb06ef7528"),
                  None],
        "data": [bytes.fromhex("0000000000000000000000000000000000000000000000000000000000000000"),
                bytes.fromhex("000000000000000000000000000000000000000000000000000002e3070e")],
        "chain_id": [1, 1],
        "n_data_bytes": [32, 32],
        "num_indexed_args": [3, 3],
        "full_signature": ["event Transfer(address indexed from, address indexed to, uint256 value)",
                           "event Transfer(address indexed from, address indexed to, uint256 value)"],
        "name": ["Transfer", "Transfer"],
        "anonymous": [False, False],
        "state_mutability": ["", ""],
        "id": ["0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef - event Transfer(address indexed from, address indexed to, uint256 value) - 0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
               "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef - event Transfer(address indexed from, address indexed to, uint256 value) - 0xdAC17F958D2ee523a2206206994597C13D831ec7"],
        "event_values": ['["0xeEDfF72A683058F8FF531e8c98575f920430FdC5", "0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D", "3151936770479715624"]',
                        '["0x9C0fdF1e074A8DB4b3bfC78cf4d48A5518B7a0BA", "0xA9D1e08C7793af67e9d92fe308d5697FB81d3E43", "774926382"]'],
        "event_keys": ['["from", "to", "value"]', '["from", "to", "value"]'],
        "event_json": ['[{"name":"from","index":0,"value_type":"address","value":"0xeEDfF72A683058F8FF531e8c98575f920430FdC5"},{"name":"to","index":1,"value_type":"address","value":"0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D"},{"name":"value","index":2,"value_type":"uint256","value":"3151936770479715624"}]',
                      '[{"name":"from","index":0,"value_type":"address","value":"0x9C0fdF1e074A8DB4b3bfC78cf4d48A5518B7a0BA"},{"name":"to","index":1,"value_type":"address","value":"0xA9D1e08C7793af67e9d92fe308d5697FB81d3E43"},{"name":"value","index":2,"value_type":"uint256","value":"774926382"}]']
    })

@pytest.fixture
def sample_decoded_trace_df():
    return pl.DataFrame({
        "transaction_index": [37, 220],
        "block_number": [21000000, 21000010],
        "transaction_hash": [bytes.fromhex("001b62412265ea5494cd57df808f44e565d898196fc588132990bb167935b385"),
                           bytes.fromhex("20F83F863231A8E1C97F14D1B2E9379D389C57D3BB82D265CD6E5D806378BA9B")],
        "selector": [bytes.fromhex("a9059cbb"), bytes.fromhex("DD62ED3E")],
        "action_to": [bytes.fromhex("502ed02100ea8b10f8d7fc14e0f86633ec2ddada"), bytes.fromhex("F19308F923582A6F7C465E5CE7A9DC1BEC6665B1")],
        "full_signature": ['function transfer(address recipient, uint256 amount) returns (bool)',
                           'function allowance(address owner, address spender) view returns (uint256)'],
        "name": ["transfer", "execute"],
        "input_values": ['["0x00000000d7F382c0b1631e81AC73D0Abd5f6F376", "8969060750056848641449546896815068662213637182689432334179124713602964652196"]',
                         '["0x000000003f2B113D180Ecb1457E450B9EFCAC3df", "0x000000007a250D5630b4Cf539739df2c5dacB4c6"]'],
        "input_keys": ['["recipient", "amount"]',
                       '["owner", "spender"]'],
        "input_json": ['[{"name":"recipient","index":0,"value_type":"address","value":"0x00000000d7F382c0b1631e81AC73D0Abd5f6F376"},{"name":"amount","index":1,"value_type":"uint256","value":"8969060750056848641449546896815068662213637182689432334179124713602964652196"}]',
                      '[{"name":"owner","index":0,"value_type":"address","value":"0x000000003f2B113D180Ecb1457E450B9EFCAC3df"},{"name":"spender","index":1,"value_type":"address","value":"0x000000007a250D5630b4Cf539739df2c5dacB4c6"}]'],
        "output_values": ['["true"]',
                          '["0"]'],
        "output_keys": ['[""]',
                        '[""]'],
        "output_json": ['[{"name":"","index":0,"value_type":"bool","value":"true"}]',
                       '[{"name":"","index":0,"value_type":"uint256","value":"0"}]'],
    })

def test_unnest_event_with_full_signature(sample_decoded_df):
    result = unnest_event(
        sample_decoded_df,
        full_signature='event Transfer(address indexed from, address indexed to, uint256 value)'
    )
    assert not result.is_empty()
    assert all(col in result.columns for col in ["from", "to", "value"])

def test_unnest_event_with_event_name(sample_decoded_df):
    result = unnest_event(sample_decoded_df, event_name="Transfer")
    assert not result.is_empty()

def test_unnest_event_with_wrong_event_name(sample_decoded_df):
    with pytest.raises(ValueError, match="No event found after filtering"):
        unnest_event(sample_decoded_df, event_name="WrongEvent")

def test_unnest_event_with_address(sample_decoded_df):
    result = unnest_event(
        sample_decoded_df,
        event_address="0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2",
        topic0="0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
    )
    assert not result.is_empty()

def test_unnest_trace_with_full_signature(sample_decoded_trace_df):
    result = unnest_trace(
        decoded_trace_df=sample_decoded_trace_df,
        full_signature="function transfer(address recipient, uint256 amount) returns (bool)"
    )
    assert not result.is_empty()
    assert all(col in result.columns for col in ["recipient", "amount", "output_0"])

def test_unnest_trace_with_function_name_and_action_to(sample_decoded_trace_df):
    result = unnest_trace(
        sample_decoded_trace_df,
        function_name="transfer",
        action_to="0x502ed02100ea8b10f8d7fc14e0f86633ec2ddada"
    )
    assert not result.is_empty()

def test_unnest_trace_with_wrong_function_name(sample_decoded_trace_df):
    with pytest.raises(ValueError, match="No trace found after filtering with the given parameters"):
        unnest_trace(sample_decoded_trace_df, function_name="WrongFunction")

def test_unnest_trace_with_selector_and_action_to(sample_decoded_trace_df):
    result = unnest_trace(
        sample_decoded_trace_df,
        selector="DD62ED3E",
        action_to="0xF19308F923582A6F7C465E5CE7A9DC1BEC6665B1"
    )
    assert not result.is_empty()
    assert all(col in result.columns for col in ["owner", "spender", "output_0"])



import polars as pl
import toml
from glaciers import get_config
from ._dataframe_utils import DataFrameType, to_polars, to_prefered_type

def unnest_event(
        decoded_log_df: DataFrameType,
        event_name: str = None,
        full_signature: str = None,
        event_address: str = None,
        topic0: str = None,
    ) -> DataFrameType:
    """
    Unnest an event columns, by filtering a DataFrame with the given parameters and then unnesting the event.

    Args:
        decoded_log_df (DataFrameType): The DataFrame (polars or pandas) to unnest.
        event_name (str, optional): The name of the event to filter the DataFrame.
        full_signature (str, optional): The full signature of the event to filter the DataFrame.
        event_address (str, optional): The address of the event to filter the DataFrame. event_address must be a hex string.
        topic0 (str, optional): The topic0 of the event to filter the DataFrame. topic0 must be a hex string.

    Returns:
        DataFrameType: The unnested DataFrame.

    Example:
        ```python
        unnest_event(
            decoded_log_df,
            event_name="Transfer",
            event_address="0x1234567890123456789012345678901234567890",
        ```
    """

    decoded_log_df = to_polars(decoded_log_df)
    
    filtered_df = decoded_log_df
    if event_name is not None:
        filtered_df = filtered_df.filter(pl.col("name").str.to_lowercase() == event_name.lower())
    if full_signature is not None:
        filtered_df = filtered_df.filter(pl.col("full_signature").str.to_lowercase() == full_signature.lower())
    if event_address is not None:
        address_col = toml.loads(get_config())["log_decoder"]["log_schema"]["log_alias"]["address"]
        col_types = filtered_df.select([pl.col(address_col)]).dtypes
        if col_types[0] == pl.String:
            filtered_df = filtered_df.filter(pl.col(address_col).str.to_lowercase().str.replace("0x", "") == event_address.lower().replace("0x", ""))
        elif col_types[0] == pl.Binary:
            filtered_df = filtered_df.filter(pl.col(address_col).bin.encode("hex").str.to_lowercase().str.replace("0x", "") == event_address.lower().replace("0x", ""))
        else:
            raise ValueError(f"Invalid column type for address: {col_types[0]}")
    if topic0 is not None:
        topic0_col = toml.loads(get_config())["log_decoder"]["log_schema"]["log_alias"]["topic0"]
        col_types = filtered_df.select([pl.col(topic0_col)]).dtypes
        if col_types[0] == pl.String:
            filtered_df = filtered_df.filter(pl.col(topic0_col).str.to_lowercase().str.replace("0x", "") == topic0.lower().replace("0x", ""))
        elif col_types[0] == pl.Binary:
            filtered_df = filtered_df.filter(pl.col(topic0_col).bin.encode("hex").str.to_lowercase().str.replace("0x", "") == topic0.lower().replace("0x", ""))
        else:
            raise ValueError(f"Invalid column type for topic0: {col_types[0]}")
    
    unique_event = filtered_df.select(pl.col("full_signature")).unique()
    if unique_event.height > 1:
        signatures = unique_event["full_signature"].to_list()
        raise ValueError(f"Event signature is not unique. Signatures: {signatures}")
    elif filtered_df.height == 0:
        raise ValueError("No event found after filtering with the given parameters")
    else:
        first_row = filtered_df.select(pl.col("event_json").str.json_decode()).row(0)[0]
        num_fields = len(first_row)
        value_types = []
        field_names = []
        for i in range(num_fields):
            value_types.append(first_row[i]["value_type"])
            if first_row[i]["name"] != "":
                field_names.append(first_row[i]["name"])
            else:
                field_names.append(f"field_{i}")

        unnesting_hex_string_encoding = toml.loads(get_config())["glaciers"]["unnesting_hex_string_encoding"]

        for (i, type) in enumerate(value_types):
            if type == "bool":
                filtered_df = filtered_df.with_columns(pl.col("event_values").str.json_decode().list.get(i).replace_strict({"false":False, "true":True}).cast(pl.Boolean).alias(f"{field_names[i]}"))
            elif "int" in type:
                filtered_df = filtered_df.with_columns(pl.col("event_values").str.json_decode().list.get(i).cast(pl.Float64).alias(f"{field_names[i]}"))
            elif "bytes" in type:
                if unnesting_hex_string_encoding:
                    filtered_df = filtered_df.with_columns(pl.col("event_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{field_names[i]}"))
                else:
                    filtered_df = filtered_df.with_columns(pl.col("event_values").str.json_decode().list.get(i).cast(pl.Binary).alias(f"{field_names[i]}"))
            elif type == "address":
                if unnesting_hex_string_encoding:
                    filtered_df = filtered_df.with_columns(pl.col("event_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{field_names[i]}"))
                else:
                    filtered_df = filtered_df.with_columns(pl.col("event_values").str.json_decode().list.get(i).cast(pl.Binary).alias(f"{field_names[i]}"))
            elif type == "string":
                filtered_df = filtered_df.with_columns(pl.col("event_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{field_names[i]}"))

        return to_prefered_type(filtered_df)


def unnest_trace(  
        decoded_trace_df: DataFrameType,
        function_name: str = None,
        action_to: str = None,
        selector: str = None,
        full_signature: str = None,

    ) -> DataFrameType:
    """
    Unnest a trace columns, by filtering a DataFrame with the given parameters and then unnesting the trace.

    Args:
        decoded_trace_df (DataFrameType): The DataFrame (polars or pandas) to unnest.
        function_name (str, optional): The name of the function to filter the DataFrame.
        action_to (str, optional): The action_to of the function to filter the DataFrame.
        selector (str, optional): The selector of the function to filter the DataFrame. selector must be a hex string.
        full_signature (str, optional): The full signature of the function to filter the DataFrame.

    Returns:
        DataFrameType: The unnested DataFrame.

    Example:
        ```python
        unnest_traces(
            decoded_trace_df,
            function_name="transfer",
            action_to="0x1234567890123456789012345678901234567890"
        )   
    """

    decoded_trace_df = to_polars(decoded_trace_df)
    
    filtered_df = decoded_trace_df
    if function_name is not None:
        filtered_df = filtered_df.filter(pl.col("name").str.to_lowercase() == function_name.lower())
    if full_signature is not None:
        filtered_df = filtered_df.filter(pl.col("full_signature").str.to_lowercase() == full_signature.lower())
    if action_to is not None:
        action_to_col = toml.loads(get_config())["trace_decoder"]["trace_schema"]["trace_alias"]["action_to"]
        col_types = filtered_df.select([pl.col(action_to_col)]).dtypes
        if col_types[0] == pl.String:
            filtered_df = filtered_df.filter(pl.col(action_to_col).str.to_lowercase().str.replace("0x", "") == action_to.lower().replace("0x", ""))
        elif col_types[0] == pl.Binary:
            filtered_df = filtered_df.filter(pl.col(action_to_col).bin.encode("hex").str.to_lowercase().str.replace("0x", "") == action_to.lower().replace("0x", ""))
        else:
            raise ValueError(f"Invalid column type for action_to: {col_types[0]}")
    if selector is not None:
        selector_col = toml.loads(get_config())["trace_decoder"]["trace_schema"]["trace_alias"]["selector"]
        col_types = filtered_df.select([pl.col(selector_col)]).dtypes
        if col_types[0] == pl.String:
            filtered_df = filtered_df.filter(pl.col(selector_col).str.to_lowercase().str.replace("0x", "") == selector.lower().replace("0x", ""))
        elif col_types[0] == pl.Binary:
            filtered_df = filtered_df.filter(pl.col(selector_col).bin.encode("hex").str.to_lowercase().str.replace("0x", "") == selector.lower().replace("0x", ""))
        else:
            raise ValueError(f"Invalid column type for selector: {col_types[0]}")
    
    unique_trace = filtered_df.select(pl.col("full_signature")).unique()
    if unique_trace.height > 1:
        signatures = unique_trace["full_signature"].to_list()
        raise ValueError(f"Trace signature is not unique. Signatures: {signatures}")
    elif filtered_df.height == 0:
        raise ValueError("No trace found after filtering with the given parameters")
    else:
        unnesting_hex_string_encoding = toml.loads(get_config())["glaciers"]["unnesting_hex_string_encoding"]

        input_first_row = filtered_df.select(pl.col("input_json").str.json_decode()).row(0)[0]
        input_num_fields = len(input_first_row)
        input_value_types = []
        input_field_names = []
        for i in range(input_num_fields):
            input_value_types.append(input_first_row[i]["value_type"])
            if input_first_row[i]["name"] != "":
                input_field_names.append(input_first_row[i]["name"])
            else:
                input_field_names.append(f"input_{i}")
        
        for (i, type) in enumerate(input_value_types):
            if type == "bool":
                filtered_df = filtered_df.with_columns(pl.col("input_values").str.json_decode().list.get(i).replace_strict({"false":False, "true":True}).cast(pl.Boolean).alias(f"{input_field_names[i]}"))
            elif "int" in type:
                filtered_df = filtered_df.with_columns(pl.col("input_values").str.json_decode().list.get(i).cast(pl.Float64).alias(f"{input_field_names[i]}"))
            elif "bytes" in type:
                if unnesting_hex_string_encoding:
                    filtered_df = filtered_df.with_columns(pl.col("input_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{input_field_names[i]}"))
                else:
                    filtered_df = filtered_df.with_columns(pl.col("input_values").str.json_decode().list.get(i).cast(pl.Binary).alias(f"{input_field_names[i]}"))
            elif type == "address": 
                if unnesting_hex_string_encoding:
                    filtered_df = filtered_df.with_columns(pl.col("input_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{input_field_names[i]}"))
                else:
                    filtered_df = filtered_df.with_columns(pl.col("input_values").str.json_decode().list.get(i).cast(pl.Binary).alias(f"{input_field_names[i]}"))
            elif type == "string":
                filtered_df = filtered_df.with_columns(pl.col("input_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{input_field_names[i]}"))

        output_first_row = filtered_df.select(pl.col("output_json").str.json_decode()).row(0)[0]
        output_num_fields = len(output_first_row)
        output_value_types = []
        output_field_names = []
        for i in range(output_num_fields):
            output_value_types.append(output_first_row[i]["value_type"])
            if output_first_row[i]["name"] != "":
                output_field_names.append(output_first_row[i]["name"])
            else:
                output_field_names.append(f"output_{i}")
        
        for (i, type) in enumerate(output_value_types):
            if type == "bool":
                filtered_df = filtered_df.with_columns(pl.col("output_values").str.json_decode().list.get(i).replace_strict({"false":False, "true":True}).cast(pl.Boolean).alias(f"{output_field_names[i]}"))
            elif "int" in type:
                filtered_df = filtered_df.with_columns(pl.col("output_values").str.json_decode().list.get(i).cast(pl.Float64).alias(f"{output_field_names[i]}"))
            elif "bytes" in type:
                if unnesting_hex_string_encoding:
                    filtered_df = filtered_df.with_columns(pl.col("output_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{output_field_names[i]}"))
                else:
                    filtered_df = filtered_df.with_columns(pl.col("output_values").str.json_decode().list.get(i).cast(pl.Binary).alias(f"{output_field_names[i]}"))
            elif type == "address":
                if unnesting_hex_string_encoding:
                    filtered_df = filtered_df.with_columns(pl.col("output_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{output_field_names[i]}"))
                else:
                    filtered_df = filtered_df.with_columns(pl.col("output_values").str.json_decode().list.get(i).cast(pl.Binary).alias(f"{output_field_names[i]}"))
            elif type == "string":
                filtered_df = filtered_df.with_columns(pl.col("output_values").str.json_decode().list.get(i).cast(pl.String).alias(f"{output_field_names[i]}"))
    
        return to_prefered_type(filtered_df)